package mimuw.mmf.clusterers.hierarchical;

import java.util.ArrayList;
import java.util.List;

import mimuw.mmf.finders.MMFRuntimeException;
import mimuw.mmf.util.Pair;

/**
 * HierarchicalClusterer is a generic class that can cluster objects of any type, if only 
 * an object able to compare the given ones is set in the constructor. 
 * 
 * @author milosz
 *
 * @param <T> type of objects that shall be clustered
 */
public class HierarchicalClusterer <T> {

	public enum THRESHOLD_TYPE {absolute, relative};
	// current list of clusters (updated at each iteration)
	private final List<HierarchicalCluster<T>> clusters;
	// current list of distances between clusters (updated at each iteration)
	private final List<List<Double>> listsOfDistances;

	private double THRESHOLD;
	private boolean stop = false;
	
	
	
	/**
	 * Constructor
	 * 
	 * @param objects list of objects of type T to cluster
	 * @param distanceCounter object that will count the distance between objects of type T
	 */
	public HierarchicalClusterer (
			List<T> objects, 
			List<List<Double>> distances,
			double threshold)
	{	
		THRESHOLD = threshold;
		int size = objects.size();
		clusters = new ArrayList<HierarchicalCluster<T>>(size); 
		for (T object : objects){
			HierarchicalCluster<T> leaf = new Leaf<T>(object);
			clusters.add(leaf);
		}
		listsOfDistances = distances;
	}



	/**
	 * Merges two clusters at the specified indices. These two clusters are removed from 
	 * the <code>clusters</code> list, and a new one, merging these two, is added at the
	 * end of the list. All distances are updated.
	 * 
	 * @param firstIndex index of the first cluster
	 * @param secondIndex index of the second cluster
	 */
	private void merge(int firstIndex, int secondIndex){

		int min = Math.min(firstIndex, secondIndex);
		int max = Math.max(firstIndex, secondIndex);

		HierarchicalCluster<T> cluster1 = clusters.get(min);
		HierarchicalCluster<T> cluster2 = clusters.get(max);

		double dist = listsOfDistances.get(min).get(max);
		//Logger.debug("Merging clusters no. %d(%s) and no. %d(%s).\n", min, cluster1.valsToString(), max, cluster2.valsToString());

		HierarchicalCluster<T> newCluster = new HierarchicalCluster<T>(cluster1, cluster2, dist);
		List<Double> distancesFromNew = new ArrayList<Double>(clusters.size());

		int current = 0;
		for (List<Double> l : listsOfDistances){
			if (current == min || current == max){
				current ++;
				continue;
			}
			Double newDistance = distanceToMerged (
					clusters.get(current), 
					cluster1,
					cluster2, 
					l.get(min), 
					l.get(max));

			l.remove(max);
			l.remove(min);
			l.add(newDistance);
			
			distancesFromNew.add(newDistance);
			current++;
		}

		// at the end - distance to the same element (has to be zero)
		distancesFromNew.add(0.0);
		listsOfDistances.remove(max);
		listsOfDistances.remove(min);
		listsOfDistances.add(distancesFromNew);

		// remove the two old clusters, and add the new one
		clusters.remove(max);
		clusters.remove(min);
		clusters.add(newCluster);
	}


	
	/**
	 * Finds the pair of clusters with the smallest distance
	 * 
	 * @return pair of their current indexes
	 */
	private Pair<Integer, Integer> findClosestClusters(){

		if (clusters.size() < 2)
			throw new MMFRuntimeException("Nothing to cluster");

		double min = Double.MAX_VALUE;
		int firstMinIndex = 0;
		int secondMinIndex = 0;
		int firstIndex = 0;
		
		// find the smallest distance
		for (List<Double> dlist : listsOfDistances){
			int secondIndex = 0;
			for (Double dist : dlist){
				if (secondIndex == firstIndex){
					secondIndex ++;
					continue;
				}
				//Logger.debug("Dist between: %s and %s is: %f\n", clustVals(firstIndex), clustVals(secondIndex), dist);
				if (dist < min){
					firstMinIndex = firstIndex;
					secondMinIndex = secondIndex;
					min = dist;
				}
				secondIndex ++;
			}
			firstIndex++;
		}
		if (min > THRESHOLD)
			stop = true;
		return new Pair<Integer, Integer>(firstMinIndex, secondMinIndex);
	}

	
	
	/**
	 * Counts the distance between cluster <code>c</code> and the cluster that 
	 * will appear as a result of merging two given clusters 
	 * 
	 * @param c cluster from which the distance will be measured
	 * @param firstClusterToMerge first from the two clusters to merge
	 * @param secondClusterToMerge second from the two clusters to merge
	 * @param distanceToFirst distance between c and firstClusterToMerge
	 * @param distanceToSecond distance between c and secondClusterToMerge
	 * @return distance between c and the new cluster
	 */
	private Double distanceToMerged (
			HierarchicalCluster<T> c, 
			HierarchicalCluster<T> firstClusterToMerge, 
			HierarchicalCluster<T> secondClusterToMerge, 
			double distanceToFirst, 
			double distanceToSecond)
	{
		int size1 = firstClusterToMerge.size();
		int size2 = secondClusterToMerge.size();
		double result = (size1 * distanceToFirst + size2 * distanceToSecond) / (size1 + size2);
		return result;
	}



	/**
	 * Perform clustering until the stop condition is reached. The stop condition is fulfilled
	 * if and only if the distance between the closest clusters is greater than THRESHOLD
	 */	
	public void performClustering(){
		while (clusters.size() > 1 && !stop){
			Pair<Integer, Integer> closest = this.findClosestClusters();
			if (stop) 
				return;
			this.merge(closest.first(), closest.second());
		}
	}

	
	/**
	 * Perform clustering until the end, when only one cluster is left. 
	 */
	public HierarchicalCluster<T> clusterToTheEnd(){
		while (clusters.size() > 1){
			Pair<Integer, Integer> closest = this.findClosestClusters();
			this.merge(closest.first(), closest.second());
		}
		return clusters.get(0);
	}
	

	/**
	 * Getter method will copy to make it impossible to change the proper list of clusters
	 * @return
	 */
	public List<HierarchicalCluster<T>> getClusters() {
		return new ArrayList<HierarchicalCluster<T>>(clusters);
	}



	public double threshold() {
		return THRESHOLD;
	}
	
}
