package mimuw.mmf.config;

import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;

import mimuw.mmf.MMF.CLUSTERING_TYPE;
import mimuw.mmf.MMF.DISTRIBUTION_DISTANCE_FUNCTION;
import mimuw.mmf.MMF.DISTRIBUTION_TYPE;
import mimuw.mmf.clusterers.hierarchical.HierarchicalClusterer.THRESHOLD_TYPE;
import mimuw.mmf.finders.MMFRuntimeException;
import mimuw.mmf.util.Constants;
import mimuw.mmf.util.Logger;
import mimuw.mmf.util.Logger.LOG_LEVEL;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * ConfigParser is useful for parsing configuration file (config.xml) in SAX mode.
 * 
 * @author milosz
 *
 */
public class ConfigSAXParser extends DefaultHandler {

	private ConfigData configData;
	
	Map<String, String> paramMap;
	String findersPath;
	Integer motifLength;
	Boolean bothStrands; 
	Integer motifsReported;
	String userMotifsFile;
	String userDatabase;
	boolean weederOnlyBestHalf;
	String tmpResultsPath;
	String verboseHelpPath;
	DISTRIBUTION_DISTANCE_FUNCTION function;
	DISTRIBUTION_TYPE type;
	Double refinementThreshold;
	CLUSTERING_TYPE clusteringType;
	double clusteringThresholdValue;
	THRESHOLD_TYPE clusteringThresholdType;
	double icThreshold;
	DISTRIBUTION_DISTANCE_FUNCTION consensusFunction;
	Integer alignmentSize;
	double gcContent;
	String species;
	LOG_LEVEL logLevel;
	boolean logPspmMotifs;
	boolean logPspmConsensus;
	JasparData jasparData;
	
	public static ConfigData parse(String fileName) throws IOException, SAXException, ParserConfigurationException {
		XMLReader xr = XMLReaderFactory.createXMLReader();
		ConfigSAXParser handler = new ConfigSAXParser();
		xr.setContentHandler(handler);
		xr.setErrorHandler(handler);
		FileReader r = new FileReader(fileName);
		//xr.setFeature("http://xml.org/sax/features/validation", true)
		xr.setFeature("http://apache.org/xml/features/validation/schema", true);

		xr.parse(new InputSource(r));
	    return handler.configData;
	}

	public ConfigSAXParser(){
		super();
		paramMap = new HashMap<String, String>();
	}
	
	@Override
	public void characters(char[] buffer, int start, int length) {}

	private Integer toInteger(String localName, Attributes attributes, String attName) throws MMFRuntimeException {
		try {
			return new Integer(attributes.getValue(attName));
		} catch (NumberFormatException e){
			throw new MMFRuntimeException("Attribute " + attName + " in element "+ localName + " has to be integer");
		}
	}
	
	private Double toDouble(String localName, Attributes attributes, String attName) throws MMFRuntimeException {
		try {
			return new Double(attributes.getValue(attName));
		} catch (NumberFormatException e){
			throw new MMFRuntimeException("Attribute " + attName + " in element "+ localName + " has to be double");
		}
	}
	
	@Override
	public void startElement(String uri, String localName, String qName, Attributes attributes){
		
		if (localName.equals("finders")){
			findersPath = attributes.getValue("path");
			if (attributes.getValue("motif_length") != null)
				motifLength = toInteger(localName, attributes, "motif_length");
			if (attributes.getValue("both_strands") != null)
				bothStrands = new Boolean(attributes.getValue("both_strands"));
			if (attributes.getValue("motifs_reported") != null)
				motifsReported = toInteger(localName, attributes, "motifs_reported");
			if (attributes.getValue("weeder_only_best_half") != null)
				weederOnlyBestHalf = new Boolean(attributes.getValue("weeder_only_best_half"));
		} else if (localName.equals("program")){
			parseNewMotifDiscoveryProgram(attributes);
		} else if (localName.equals("background")) {
			species = attributes.getValue("species");
			gcContent = toDouble(localName, attributes, "gcContent");
		}  else if (localName.equals("refine")){
			refinementThreshold = toDouble(localName, attributes, "threshold");
		} else if (localName.equals("paths")){
			tmpResultsPath = attributes.getValue("tmp_results");
			verboseHelpPath = attributes.getValue("verbose_help");
			userMotifsFile = attributes.getValue("userMotifsFile");
			userDatabase = attributes.getValue("userDatabase");
		} else if (localName.equals("clustering")){
			clusteringType = CLUSTERING_TYPE.valueOf(attributes.getValue("type"));
			clusteringThresholdType = THRESHOLD_TYPE.valueOf(attributes.getValue("thresholdType"));
			clusteringThresholdValue = toDouble(localName, attributes, "thresholdValue");
		} else if (localName.equals("consensus")){
			consensusFunction = DISTRIBUTION_DISTANCE_FUNCTION.valueOf(attributes.getValue("function"));
			icThreshold = toDouble(localName, attributes, "icThreshold");
			String alignmentSizeString = attributes.getValue("alignmentSize");
			if (null != alignmentSizeString)
				alignmentSize = toInteger(localName, attributes, "alignmentSize");
		} else if (localName.equals("distribution")){
			function = DISTRIBUTION_DISTANCE_FUNCTION.valueOf(attributes.getValue("function"));
			type = DISTRIBUTION_TYPE.valueOf(attributes.getValue("type"));
		} else if (localName.equals("logs")){
			logLevel = LOG_LEVEL.valueOf(attributes.getValue("level"));
			logPspmMotifs = new Boolean(attributes.getValue("pspmMotif"));
			logPspmConsensus = new Boolean(attributes.getValue("pspmConsensus"));
		} else if (localName.equals("jaspar")){
			jasparData = new JasparData(attributes);
		}
	}
	
	private void parseNewMotifDiscoveryProgram(Attributes attributes){
		String program = attributes.getValue("name");
		String params = attributes.getValue("params");
		paramMap.put(program, params==null ? "" : params);
	}
	
	
	@Override
	public void endElement(String uri, String localName, String qName){}

	
	public void endDocument()  throws SAXException {
		
		configData = new ConfigData(
				paramMap, findersPath, motifLength, bothStrands, motifsReported,
				weederOnlyBestHalf, 
				tmpResultsPath, verboseHelpPath, userMotifsFile, userDatabase,
				function, type,
				refinementThreshold,
				clusteringType,	clusteringThresholdType, clusteringThresholdValue,
				icThreshold, consensusFunction, alignmentSize,
				species, gcContent,
				logLevel, logPspmMotifs, logPspmConsensus,
				jasparData);
	}
	
	@Override
	public void warning(SAXParseException exception) {
		Logger.err("WARNING: line %d: %s\n", exception.getLineNumber(), exception.getMessage());
	}

	@Override
	public void error(SAXParseException e) throws SAXException{
		Logger.err("Error in line %d of %s: %s\n", e.getLineNumber(), Constants.CONFIG_PATH, e.getMessage());
		throw e;
	}

	@Override
	public void fatalError(SAXParseException exception) throws SAXException {
		Logger.err("Fatal error in line %d of %s: %s\n", exception.getLineNumber(), Constants.CONFIG_PATH, exception.getMessage());
		throw (exception);
	}
}
