package mimuw.mmf.finders.mdscan;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import mimuw.mmf.WrongParameterException;
import mimuw.mmf.finders.AbstractMotifFinder;
import mimuw.mmf.finders.AbstractResultsParser;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.finders.WrongValueReturnedException;
import mimuw.mmf.motif.Motif;
import mimuw.mmf.util.Constants;
import mimuw.mmf.util.Logger;
import mimuw.mmf.util.StreamGobbler;

import org.biojavax.bio.seq.RichSequence;

public class MDscanFinder extends AbstractMotifFinder {

	public static final String MDSCAN_PATH = Constants.FINDERS_PATH + "MDscan/MDscan.linux";
	public static final String IDENTIFIER = "mdscan";

	public MDscanFinder(){
		super();
	}

	public List<Motif> findMotifs(File seqFile, Map<String, RichSequence> seqMap)
		throws WrongValueReturnedException, MMFException 
	{
		Logger.trace(">>FinderMDscan.run()");
		List<Motif> result = null;
		StringBuffer cmd = new StringBuffer(100);
		String tmpFile = Constants.TMP_RESULTS_DIR + "results.mdscan";
		Runtime rt = Runtime.getRuntime();
		Process proc;
		
		putParameter("i", seqFile.getAbsolutePath());
		putParameter("o", tmpFile);
		putParameter("g", "1");  // don't want to see messages during the run
		
		cmd .append(MDSCAN_PATH)
			.append(getParametersString());

		try {
			Logger.info("MDscanFinder.findMotifs() - command: \n\t%s\n", cmd);
			proc = rt.exec(cmd.toString());
			StreamGobbler errorGobbler = new StreamGobbler(proc.getErrorStream(), "ERROR");
			StreamGobbler outputGobbler = new StreamGobbler(proc.getInputStream(), "OUTPUT");
			errorGobbler.start();
			outputGobbler.start();
			//OutputStream os = proc.getOutputStream();
			int exitVal = proc.waitFor();
			Logger.trace("ExitValue: %d\n", exitVal);
			if (exitVal != 0)
				throw new WrongValueReturnedException(exitVal, getProgramIdentifier());
			AbstractResultsParser parser = new MDscanResultsParser(tmpFile);
			result = parser.parseFile(seqMap);
		} catch (IOException e) {
			throw new MDscanException(e);
		} catch (InterruptedException e) {
			throw new MDscanException(e);
		}
		Logger.trace("<<FinderMDscan.run()");
		return result;
	}


	public String getProgramIdentifier() {
		return IDENTIFIER;
	}

	@Override
	public <T> void setParameter(T name, String val) throws WrongParameterException {

		if (name.toString().length() != 1)
			throw new WrongParameterException("MDscan: Parameter name of length 1 required for each parameter (found " + name + ")");

		switch(name.toString().charAt(0)){
		case 'w':
			checkParameterRange(name, val, MIN_MOTIF_LEN, MAX_MOTIF_LEN);
			break;
		case 'o':
		case 'f':
		case 'b':
		case 'g':
			throw new WrongParameterException("Parameter -" + name + " is forbidden in the parameters file");
		case 't':
			checkParameterRange(name, val, 1, 1000);
			break;
		case 'c':
			checkParameterRange(name, val, 1, 1000);
			break;
		case 'e':	// TODO: what does this parameter exactly mean?
			break;
		case 's':
			checkParameterRange(name, val, 1, 1000);
			break;
		case 'r':
			checkParameterRange(name, val, MIN_REPORTED_MOTIFS_NUMBER, MAX_REPORTED_MOTIFS_NUMBER);
			break;
		case 'n':
			checkParameterRange(name, val, 1, 100);
			break;
		}
		putParameter(name, val);
	}

	
	
	@Override
	public void setParametersAsSingleString(String pS) throws WrongParameterException {
		String[] tokens = pS.split("-");
		for (String token: tokens){
			token = token.trim();
			if (token.equals(""))
				continue;
			String keyVal[] = token.split("\\s", 2);
			String name = keyVal[0];
			if (name.length() > 1)
				throw new WrongParameterException("MDscan parameter's name cannot be longer than 1 (" + name + ")");
			String val = keyVal.length == 2 ? keyVal[1] : null;
			setParameter(name, val);
		}
	}

	public void setMotifLength(int len) throws WrongParameterException {
		setParameter("w", "" + len);
	}

	public void setSearchBothStrands(boolean both) throws WrongParameterException{
		if (!both)
			throw new WrongParameterException("MDscan always looks for motifs ON BOTH strands");
	}
	
	public void setDesiredNumberOfMotifs(int n) throws WrongParameterException {
		setParameter('r', ""+n);
	}
}
