package mimuw.mmf.finders.meme;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import org.biojavax.bio.seq.RichSequence;

import mimuw.mmf.WrongParameterException;
import mimuw.mmf.finders.AbstractMotifFinder;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.finders.WrongValueReturnedException;
import mimuw.mmf.motif.Motif;
import mimuw.mmf.util.Constants;
import mimuw.mmf.util.Logger;
import mimuw.mmf.util.StreamGobbler;

public class MEMEFinder extends AbstractMotifFinder {

	private static final String MEME_PATH = Constants.FINDERS_PATH + "meme/bin/meme";
	public static final String IDENTIFIER = "meme";
	
	public MEMEFinder(){
		super();
	}
	
	
	
	public List<Motif> findMotifs (File seqFile, Map<String, RichSequence> seqMap) 
		throws WrongValueReturnedException, MMFException 
	{
		Logger.trace(">>MEMEFinder.run()\n");
		
		List<Motif> result;
		StringBuffer cmd = new StringBuffer(100);
		String tmpResultsFile = Constants.TMP_RESULTS_DIR + "results.meme";
		Runtime rt = Runtime.getRuntime();
		Process proc;
		
		putParameter("dna", null);
		putParameter("text", null);
		putParameter("nostatus", null);
		
		cmd .append(MEME_PATH + ' ' + seqFile + getParametersString());

		try {
			Logger.info("MEMEFinder.findMotifs() - command: \n\t%s\n", cmd);
			proc = rt.exec(cmd.toString());
			FileOutputStream fos = new FileOutputStream(tmpResultsFile);
			StreamGobbler errorGobbler = new StreamGobbler(proc.getErrorStream(), "ERROR");
			StreamGobbler outputGobbler = new StreamGobbler(proc.getInputStream(), "OUTPUT", fos);
			errorGobbler.start();
			outputGobbler.start();
			int exitVal = proc.waitFor();
			errorGobbler.join();
			outputGobbler.join();
			if (exitVal != 0)
				throw new WrongValueReturnedException(exitVal, getProgramIdentifier());
			
			MEMEResultsParser parser = new MEMEResultsParser(tmpResultsFile);
			result = parser.parseFile(seqMap);
		} catch (IOException e) {
			throw new MEMEException(e);
		} catch (InterruptedException e) {
			throw new MEMEException(e);
		}
		Logger.trace("<<MEMEFinder.run()\n");
		return result;
	}


	public String getProgramIdentifier() {
		return IDENTIFIER;
	}
	
	@Override
	public <T> void setParameter(T name, String val) throws WrongParameterException {

		if (name.equals("h")){
			throw new WrongParameterException("Parameter -h is not accepted here");
		} else if (name.equals("dna") || name.equals("protein")){
			if (val != null)
				throw new WrongParameterException("Unexpected value ("+val+") for parameter "+name+" in MEME");
		} else if (name.equals("mod")){
			if (!val.equals("oops") && !val.equals("zoops") && !val.equals("anr"))
				throw new WrongParameterException("Unexpected value ("+val+") for parameter "+name+" in MEME");
		} else if (name.equals("nmotifs")){
			checkParameterRange(name, val, MIN_REPORTED_MOTIFS_NUMBER, MAX_REPORTED_MOTIFS_NUMBER);
		} else if (name.equals("evt")){
			checkParameterPositiveFloat(name, val);
		} else if (name.equals("nsites") || name.equals("minsites") || name.equals("maxsites")){
			checkParameterRange(name, val, 1, 500);
		} else if (name.equals("wnsites")){
			checkFloatParameterRange(name, val, new Float(0), new Float(1));
		} else if (name.equals("w") || name.equals("minw") || name.equals("maxw")){
			checkParameterRange(name, val, MIN_MOTIF_LEN, MAX_MOTIF_LEN);
		} else if (name.equals("nomatrim")){
			if (val != null)
				throw new WrongParameterException("Unexpected value ("+val+") for parameter "+name+" in MEME");
		} else if (name.equals("wg") || name.equals("ws")){
			checkFloatParameterRange(name, val, Float.MIN_VALUE, Float.MAX_VALUE);
		} else if (name.equals("noendgaps") || name.equals("revcomp")|| name.equals("pal")){
		} else if (name.equals("bfile") || name.equals("plib")){	
		} else if (name.equals("maxiter")){
			checkParameterRange(name, val, 1, 1000);
		} else if (name.equals("distance")){
			checkFloatParameterRange(name, val, new Float(0), Float.MAX_VALUE);
		} else if (name.equals("prior")){
			if (!val.equals("dirichlet") && !val.equals("dmix") && !val.equals("mega") && !val.equals("megap") && !val.equals("addone"))
				throw new WrongParameterException("Unexpected value ("+val+") for parameter "+name+" in MEME");;
		} else if (name.equals("b")){
			checkFloatParameterRange(name, val, new Float(0), Float.MAX_VALUE);
		} else if (name.equals("spfuzz")){
			checkParameterPositiveFloat(name, val);
		} else if (name.equals("spmap")){
			if (!val.equals("uni") && !val.equals("pam"))
				throw new WrongParameterException("Unexpected value ("+val+") for parameter "+name+" in MEME");
		} else if (name.equals("cons")){
		} else if (name.equals("text") || name.equals("nostatus")){
		} else if (name.equals("maxsize")){
			checkParameterRange(name, val, 1, Integer.MAX_VALUE);
		} else if (name.equals("p")){
			checkParameterRange(name, val, 1, 256);
		} else if (name.equals("time")){
			checkParameterRange(name, val, 1, Integer.MAX_VALUE);
		} else if (name.equals("sf")){
		} else 
			throw new WrongParameterException("Parameter unknown by MEME: " + name);
		
		putParameter(name, val);
	}



	@Override
	public void setParametersAsSingleString(String pS) throws WrongParameterException {
		String[] tokens = pS.split("-");
		for (String token: tokens){
			token = token.trim();
			if (token.equals(""))
				continue;
			String keyVal[] = token.split("\\s", 2);
			String name = keyVal[0];
			String val = keyVal.length == 2 ? keyVal[1] : null;
			setParameter(name, val);
		}
	}



	public void setMotifLength(int len) throws WrongParameterException {
		setParameter("w", "" + len);
	}



	public void setSearchBothStrands(boolean both) {
		if (both)
			putParameter("revcomp", null);
		else
			removeParameter("revcomp");
	}



	public void setDesiredNumberOfMotifs(int n) throws WrongParameterException {
		if (n < MIN_REPORTED_MOTIFS_NUMBER || n > MAX_REPORTED_MOTIFS_NUMBER)
			throw new WrongParameterException("Wrong number of motifs: " + n);
		putParameter("nmotifs", "" + n);
	}
	
}