package mimuw.mmf.finders.weeder;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.biojavax.bio.seq.RichSequence;

import mimuw.mmf.finders.AbstractResultsParser;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.motif.Motif;
import mimuw.mmf.motif.MotifOccurrence;
import mimuw.mmf.motif.OccurrenceBeyondSequenceException;
import mimuw.mmf.util.Constants;
import mimuw.mmf.util.Logger;

public class WeederResultsParser extends AbstractResultsParser {


	List<String> seqNames;

	public WeederResultsParser(String resultsFile) throws MMFException {
		super(resultsFile);
	}

	/**
	 * Parses one motif from file. Assumes that when called, variable 'line' contains the first 
	 * motif sequence (pattern)
	 */
	@Override
	protected Motif parseMotif(Map<String, RichSequence> seqMap) {

		int length = line.trim().length();

		this.readUntilRegex("\\d\\sredundant\\smotifs.*");
		readln();
		String [] subSeqs = line.trim().split("-");


		List<MotifOccurrence> occurrences = new LinkedList<MotifOccurrence>();

		this.readUntilRegex("Seq\\s+St\\s+oligo\\s+pos\\s+match");
		// parse occurrences
		while (readln().trim().matches("\\d.*")){
			line = line.trim();
			String tokens [] = line.split("\\s+");
			if (!line.matches("\\d+\\s+[\\+\\-]\\s+\\S+\\s+\\d+\\s+\\(\\d+\\.\\d{2}\\)"))
				wrongFormat();
			if (tokens.length < 5)
				wrongFormat();
			int seqIndex = new Integer(tokens[0]);
			if (!tokens[1].matches("[\\+\\-]"))
				wrongFormat("+ or - ", tokens[1]);
			boolean forward = tokens[1].matches("\\+");
			boolean weakOccurrence = tokens[2].trim().charAt(0) == '[';

			if (! (weakOccurrence && Constants.WEEDER_CONSIDER_ONLY_BEST_HALF)){
				int position = new Integer(tokens[3]);
				try {
					MotifOccurrence o = new MotifOccurrence(
							seqMap.get(getSeqName(seqIndex, seqNames)), 
							position, forward, length);
					occurrences.add(o);
				} catch (OccurrenceBeyondSequenceException e) {
					Logger.warn("WARNING: %s. \n  Occurrence removed.\n", e.getMessage());
				}
			}
		}
		readUntilRegex("={42}");	// end of motif's description
		WeederMotif motif = new WeederMotif(length, occurrences, WeederFinder.IDENTIFIER, motifIndex, Arrays.asList(subSeqs));
		this.motifIndex ++;
		return motif;
	}



	@Override
	public List<Motif> parseFile(Map<String, RichSequence> seqMap) {

		boolean end = false;
		seqNames = this.parseSequenceNumbering();
		Logger.debug("Found %d sequences\n", seqNames.size());
		List<Motif> motifsList = new ArrayList<Motif>();

		readUntilRegex("(Sorry.*)|(\\*\\*\\* Interesting.*)");
		if (line.trim().startsWith("Sorry"))	// no motifs found
			return motifsList;
		
		for (;;){
			readln();
			if (line == null){
				end = true;
				break;
			} else if (!line.matches("\\S+.*"))
				this.readUntilRegex("\\S+");
			Motif motif = parseMotif(seqMap);
			motifsList.add(motif);
			readln();
			if (line == null){
				end = true;
				break;
			}
			readln();
			if (line.trim().startsWith("*** Interesting motifs (not highest-ranking)"))
				break;
		}

		Logger.info("Highest ranking motifs: %d\n", motifsList.size());
		//this.readUntilRegex("\\*\\*\\* Interesting.*");
		if (!end){
			readln();

			if (line != null){
				for (;;){
					this.readUntilRegex("\\S+");
					Motif motif = parseMotif(seqMap);
					motifsList.add(motif);
					readln();
					if (line == null)
						break;
				}
			}
		}

		//BioTools.addWeigthsToMotifs(motifsList);
		return motifsList;
	}



	/**
	 * Reads through the section describing sequences read from fasta file
	 * The section has the following format (example):
	 * 
	 * Your sequences:
	 * Sequence 1 : >First_seq_name
	 * Sequence 2 : >Second_seq_name
	 * Sequence 3 : >Third_seq_name
	 * 
	 * @return list of the sequences' names
	 */
	private List<String> parseSequenceNumbering() {

		List<String> seqNamesLists = new ArrayList<String>();
		int seq = 0;

		this.readUntilRegex("Sequence\\s\\d.*");

		while (line.trim().startsWith("Sequence")){
			seq ++;
			String seqName = line.split(">")[1].trim().split("\\s")[0];
			seqNamesLists.add(seqName);
			Logger.debug("Sequence read: %s\n", seqName);
			this.readln();
		}
		return seqNamesLists;
	}



	private String getSeqName(int seqIndex, List<String> seqNamesList){
		return seqNamesList.get(seqIndex - 1);
	}



	private void readUntilRegex(String regex) {
		while (readln() != null && !line.trim().matches(regex)){
		}
		if (line == null)
			wrongFormat("regex: " + regex, "end of file");	
	}

}
