package mimuw.mmf.motif;

import mimuw.mmf.clusterers.Weighable;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.finders.MMFRuntimeException;
import mimuw.mmf.util.BioTools;
import mimuw.mmf.util.QuickBioTools;

import org.biojava.bio.BioException;
import org.biojava.bio.dist.Count;
import org.biojava.bio.dist.Distribution;
import org.biojava.bio.dist.DistributionTools;
import org.biojava.bio.dist.SimpleDistribution;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.FiniteAlphabet;

public abstract class AbstractMotif implements Weighable  {

	protected int length;
	private Count[] counts;
	private Distribution[] distributions;

	private double [][] pspm;
	private double [][] revCompPspm;
	private double avgInfo;
	private String consensus;

	private double [] fittingProbabilities;
	private double [] fittingProbabilitiesRevComp;
	private double sumOfFittingProbabilities = 0;
	private double sumOfFittingProbabilitiesRevComp = 0;

	public AbstractMotif() {
		super();
	}
	
	public AbstractMotif(Count[] counts, Distribution bg) {
		super();
		init(counts, bg);
	}
	
	protected void init(Count[] counts, Distribution bg){
		try {
			this.length = counts.length;
			this.counts = counts;
			this.distributions = new Distribution[length()];
			for (int i = 0; i < distributions.length; i++){
				distributions[i] = (SimpleDistribution) BioTools.applyLaplaceCorrection(counts[i], bg);
			}
			this.avgInfo = countAverageInformation();
			pspm = QuickBioTools.toPSPMArray(distributions);
			revCompPspm = QuickBioTools.toPSPMArrayRevComp(distributions);
			consensus = BioTools.getConsensusSequence(this.distributions);
		} catch (BioException e) {
			throw new MMFRuntimeException(e);
		}	
	}
	

	public abstract String origin();

	public int length(){
		return length;
	}

	public abstract double getWeight();

	public Count[] getCounts() {	return counts;	}

	public Distribution[] getDistributions() { return distributions; }

	public String consensus() { return consensus; }

	public double avgInfo(){ return avgInfo; };

	public abstract FiniteAlphabet getAlphabet();

	public double[][] getPSPM() {
		return pspm;
	}

	public double[][] getRevCompPSPM() {
		return revCompPspm;
	}

	private double countAverageInformation() {
		double avgInfo = 0;
		for (Distribution d: getDistributions())
			avgInfo += DistributionTools.bitsOfInformation(d);
		avgInfo /= getDistributions().length;
		return avgInfo;
	}

	public void prepareFittingProbabilities(Sequence concatSeq) throws MMFException {
		try {
			fittingProbabilities = QuickBioTools.getMotifProbsOnSingleSequence(getPSPM(), concatSeq);
			fittingProbabilitiesRevComp = QuickBioTools.getMotifProbsOnSingleSequence( getRevCompPSPM(), concatSeq);
			for (double p : fittingProbabilities)
				sumOfFittingProbabilities += p;
			for (double p : fittingProbabilitiesRevComp)
				sumOfFittingProbabilitiesRevComp += p;
		} catch (Exception e) {
			throw new MMFException(e);
		}
	}

	public double[] fittingProbabilities() throws NotInitializedException {
		if (fittingProbabilities == null)
			throw new NotInitializedException("Must call prepareFittingProbabilities() first");
		return fittingProbabilities;
	}

	public double getSumOfFittingProbabilities() throws NotInitializedException {
		if (fittingProbabilities == null)
			throw new NotInitializedException("Must call prepareFittingProbabilities() first");
		return sumOfFittingProbabilities;
	}

	public double getSumOfFittingProbabilitiesRevComp() throws NotInitializedException {
		if (fittingProbabilitiesRevComp == null)
			throw new NotInitializedException("Must call prepareFittingProbabilities() first");
		return sumOfFittingProbabilitiesRevComp;
	}

	public double[] getFittingProbabilitiesRevComp() throws NotInitializedException {
		if (fittingProbabilitiesRevComp == null)
			throw new NotInitializedException("Must call prepareFittingProbabilities() first");
		return fittingProbabilitiesRevComp;
	}


	double[] getFittingProbabilities() {
		return fittingProbabilities;
	}

	void setFittingProbabilities(double[] fittingProbabilities) {
		this.fittingProbabilities = fittingProbabilities;
	}
}