package mimuw.mmf.motif.external;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import mimuw.mmf.finders.FileFormatException;
import mimuw.mmf.motif.TemporaryDatabaseMotif;
import mimuw.mmf.util.BioTools;
import mimuw.mmf.util.Constants;

import org.biojava.bio.dist.Count;
import org.biojava.bio.dist.IndexedCount;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.symbol.AtomicSymbol;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.utils.ChangeVetoException;

public class ExternalMotifsReader {

	public static final String USER_MOTIFS_IDENTIFIER = "EXTERNAL_DATA";
	public static final String EXTERNAL_DATABASE_MOTIFS = "DB_EXTERNAL";
	
	public List<ExternalMotif> readUserMotifs(String filename) throws IOException
	{	
		List<ExternalMotif> motifs = new LinkedList<ExternalMotif>();
		BufferedReader bufRead = new BufferedReader(new FileReader(filename));
		List<Count> counts = new LinkedList<Count>();
		String line, tokens[];
		int lineNumber = 0, index = 0;
		
		try { 
			while ((line = bufRead.readLine()) != null){
				lineNumber ++;
				if (line.trim().equals("")){
					if (!counts.isEmpty()){
						Count[] c = new Count[counts.size()];
						counts.toArray(c);
						ExternalMotif motif = new ExternalMotif(c, Constants.getBgDist(), 
								USER_MOTIFS_IDENTIFIER, index++);
						motifs.add(motif);
						counts.clear();
					}
					continue;
				}
				tokens = line.split("\\s+");
				if (tokens.length < 4 || tokens.length > 6){
					System.err.printf("Line: #%s#, only %d tokens\n", line, tokens.length);
					throw new FileFormatException(lineNumber, filename);
				}
				// accept the header line (not obligatory)
				if (tokens.length == 5 && tokens[1].equals("A")) continue;
				if (tokens.length == 4 && tokens[0].equals("A")) continue;
				int shift = (tokens.length > 4) ? 1 : 0;
				Double a, c, g, t;
				try {
					a = Double.parseDouble(tokens[shift+0]);
					c = Double.parseDouble(tokens[shift+1]);
					g = Double.parseDouble(tokens[shift+2]);
					t = Double.parseDouble(tokens[shift+3]);
				} catch (NumberFormatException e){
					throw new FileFormatException(lineNumber, filename);
				}
				counts.add(BioTools.getDNACount(a, c, g, t));
			}
			if (!counts.isEmpty()){
				Count[] c = new Count[counts.size()];
				counts.toArray(c);
				ExternalMotif motif = new ExternalMotif(c, Constants.getBgDist(), USER_MOTIFS_IDENTIFIER, index++);
				motifs.add(motif);
				counts.clear();
			}
		} finally {
			bufRead.close();
		}
		return motifs;
	}


	public List<ExternalMotif> readExternalDatabase(String filename) 
	throws ChangeVetoException, IOException
	{
		Map<String, TemporaryDatabaseMotif>  motifs = new HashMap<String, TemporaryDatabaseMotif>();
		BufferedReader bufRead = new BufferedReader(new FileReader(filename));
		String line, tokens[], id;
		AtomicSymbol symbol;
		int pos, lineNumber = 0;
		double count;
		TemporaryDatabaseMotif m;

		try {
			while ((line = bufRead.readLine()) != null){
				lineNumber ++;
				if (line.trim().equals(""))
					continue;
				tokens = line.split("\\s");
				if (tokens.length != 4)
					throw new FileFormatException(lineNumber, filename);
				id = tokens[0];
				m = motifs.get(id);
				if (m == null){
					m = new TemporaryDatabaseMotif(id);
					motifs.put(id, m);
				}
				if (tokens[1].length() != 1)
					throw new FileFormatException(lineNumber, filename);
				try {
					symbol = (AtomicSymbol) DNATools.forSymbol(tokens[1].charAt(0));
					pos = Integer.parseInt(tokens[2]);
					count = Double.parseDouble(tokens[3]);
					m.addData(symbol, pos, count);
				} catch (IllegalSymbolException e) {
					throw new FileFormatException(lineNumber, filename);
				} catch (NumberFormatException e) {
					throw new FileFormatException(lineNumber, filename);
				}
			}
		} finally {
			bufRead.close();
		}
		
		List<ExternalMotif> result = new LinkedList<ExternalMotif>();
		int index = 0;
		for (TemporaryDatabaseMotif motif: motifs.values()){
			IndexedCount[] counts = motif.getTmpCounts().toArray(new IndexedCount[motif.getTmpCounts().size()]);
			ExternalMotif dbMotif = new ExternalMotif(counts, Constants.getBgDist(), EXTERNAL_DATABASE_MOTIFS, index++);
			result.add(dbMotif);
		}
		return result;
	}
	
}
