package mimuw.mmf.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collections;
import java.util.Formatter;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import mimuw.mmf.WrongParameterException;
import mimuw.mmf.clusterers.DistanceCounter;
import mimuw.mmf.finders.MMFException;
import mimuw.mmf.finders.MMFRuntimeException;
import mimuw.mmf.motif.AbstractMotif;
import mimuw.mmf.motif.Motif;

import org.biojava.bio.BioException;
import org.biojava.bio.dist.Count;
import org.biojava.bio.dist.Distribution;
import org.biojava.bio.dist.DistributionFactory;
import org.biojava.bio.dist.DistributionTools;
import org.biojava.bio.dist.IndexedCount;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.RNATools;
import org.biojava.bio.symbol.Alphabet;
import org.biojava.bio.symbol.AtomicSymbol;
import org.biojava.bio.symbol.FiniteAlphabet;
import org.biojava.bio.symbol.IllegalAlphabetException;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.Symbol;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.ChangeVetoException;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;

/**
 * This class contains various static methods for manipulating biological data.
 * 
 * @author milosz
 *
 */
public class BioTools {

	public static final double EPSILON = 10E-12;
	
	public static final FiniteAlphabet dna = DNATools.getDNA();
	public static final FiniteAlphabet rna = RNATools.getRNA();

	public static final AtomicSymbol a = DNATools.a();
	public static final AtomicSymbol c = DNATools.c();
	public static final AtomicSymbol g = DNATools.g();
	public static final AtomicSymbol t = DNATools.t();
	public static final AtomicSymbol u = RNATools.u();
	
	public static final AtomicSymbol[] dnaSymbols = {a, c ,g, t};
	public static final AtomicSymbol[] rnaSymbols = {a, c ,g, u};
	
	public static interface DistributionDistanceCounter extends DistanceCounter<Distribution>{
		public double distance(Distribution p, Distribution q);
	}

	@Deprecated
	public static class RelativeEntropyCounter implements DistributionDistanceCounter {

		public double distance(Distribution p, Distribution q) {
			return relativeEntropy(p, q);		
		}
	}

	@Deprecated
	public static class DpqCounter implements DistributionDistanceCounter {

		public double distance(Distribution p, Distribution q) {
			return DpqDistance(p, q);
		}
	}
	
	
	public static Count getDNACount(double a, double c, double g, double t){
		Count count = new IndexedCount(DNATools.getDNA());
		try {
			count.setCount(DNATools.a(), a);
			count.setCount(DNATools.c(), c);
			count.setCount(DNATools.g(), g);
			count.setCount(DNATools.t(), t);
		} catch (BioException e) {
			throw new MMFRuntimeException(e);
		}
		return count;
	}
	
	
	public static Distribution getDNADistribution(double gcContent){
		return getDNADistribution((1-gcContent)/2, gcContent/2, gcContent/2, (1-gcContent)/2);
	}
	
	
	public static Distribution getDNADistribution(double a, double c, double g, double t){
		Distribution d;
		try {
			double sum = a + c + g + t;
			if (sum < (1-EPSILON) | sum > (1+EPSILON))
				throw new MMFRuntimeException("Wrong sum: " + sum);
			d = DistributionFactory.DefaultDistributionFactory.DEFAULT.createDistribution(DNATools.getDNA());
			d.setWeight(DNATools.a(), a);
			d.setWeight(DNATools.c(), c);
			d.setWeight(DNATools.g(), g);
			d.setWeight(DNATools.t(), t);
			return d;
		} catch (BioException e) {
			throw new MMFRuntimeException(e);
		}
	}
	

	public static double applyPenalty(double result, int overlap, int outside){
		return (result * (2*overlap + outside))/ (2*overlap);
	}


	@Deprecated
	public static double getDistance(AbstractMotif m1, AbstractMotif m2, 
			DistributionDistanceCounter counter, Distribution background) 
	
	{	
		Distribution[] d1 = m1.getDistributions();
		Distribution[] d2 = m2.getDistributions();
		try {
			//TODO: sth makes it slower here!!!
			Distribution[] d2Rev = reverseComplement(d2);
			double d1ToD2 = getBestAvgDistance(counter, d1, d2, background);
			double d1ToD2Rev = getBestAvgDistance(counter, d1, d2Rev, background);
			return Math.min(d1ToD2, d1ToD2Rev);
		} catch (WrongParameterException e) {
			throw new MMFRuntimeException(e);
		}
	}


	@Deprecated
	public static double getBestAvgDistance (
			DistributionDistanceCounter counter, 
			Distribution[] first, 
			Distribution[] second,
			Distribution background)
	{
		double bestScore = Double.MAX_VALUE;
		int len1 = first.length;
		int len2 = second.length;
		int bestShift = -(len1-1);
		
		for (int shift = bestShift; shift < len2; shift ++){

			int overlapLength = shift < 0 ? Math.min(len1+shift, len2)	: Math.min(len1, len2 - shift);
			int outsideLength = len1 + len2 - 2*overlapLength;
			//int totalLength = outsideLength + overlapLength;

			double current = getTotalDistance(counter, first, shift, second, background);
			double average = current / (overlapLength + outsideLength);
			//TODO I've changed current to average below
			double withPenalty = applyPenalty(average, overlapLength, outsideLength);
			//System.out.printf("For shift %d: OVERLAP=%d, OUT=%d, distance=%f, +penalty=%f, avg dist=%f\n", 
			//		shift, overlapLength, outsideLength, current, withPenalty, current/totalLength);
			if (withPenalty < bestScore){
				bestScore = withPenalty;
				bestShift = shift;
			}				
		}
		Logger.info("   Best shift: %d, bestScore: %.2f ",bestShift, bestScore);
		return bestScore;
	}



	/**
	 * Counts sum of the distances between the corresponding distributions from given arrays. If <i>firstShiftedRight</i> 
	 * is greater than zero, the first array is 'padded' with backgorund distribution (exactly <i>firstShiftedRight</i> times).
	 * Otherwise, this happens to the second array. If after the padding one of the arrays is shorter (which may often happen), 
	 * it is also padded (on the right end) with background. The distance between two distributions is counted using method
	 * <code>counter.distance(..)</code>.
	 * 
	 * @param counter Object that will be used to count the distance between two distributions
	 * @param first First array with distributions
	 * @param firstShiftedRight Offset describing how much the first array should be shifted to
	 * 		the right. If negative, first array will be shifted to the left
	 * @param second Second array with distributions
	 * @param background The background distribution
	 * @return
	 */
	@Deprecated
	public static double getTotalDistance(
			DistributionDistanceCounter counter, 
			Distribution[] first, 
			int firstShiftedRight, 
			Distribution[] second,
			Distribution background)
	{
		double result = 0;
		int indexFirst = 0;
		int indexSecond = 0;
		int firstLen = first.length;
		int secondLen = second.length;

		//TODO: throw an exception if offset is longer than the size of the second distribution array

		// START: counts sum of the distances between padding and the beginning
		if (firstShiftedRight < 0){
			for (; indexFirst < -firstShiftedRight; indexFirst++){
				result += counter.distance(first[indexFirst], background);
			}
		} else {
			for (; indexSecond < firstShiftedRight; indexSecond++){
				result += counter.distance(background, second[indexSecond]);
			}
		}

		// MIDDLE: counts the main part: distances of the corresponding distributions
		while (indexFirst < firstLen && indexSecond < secondLen){
			result += counter.distance(first[indexFirst], second[indexSecond]);
			indexFirst++;
			indexSecond++;
		}

		// END: adds distances between the distributions at the end and padding in the other sequence
		while (indexFirst < firstLen){
			result += counter.distance(first[indexFirst], background);
			indexFirst++;
		}
		while (indexSecond < secondLen){
			result += counter.distance(background, second[indexSecond]);
			indexSecond++;
		}
		return result;
	}



	



	/**
	 * Applies the Laplace correction to the given count. Zeros are eliminated by
	 * adding the background distribution to each symbol, and then normalizing it.
	 * Such an obtained count is then transformed into the distribution and returned 
	 * as a result
	 * 
	 * @param count - the count to change
	 * @param background - probabilistic background description
	 * @return
	 * @throws BioException 
	 */
	public static Distribution applyLaplaceCorrection(Count count, Distribution background) throws BioException{
		Count c = new IndexedCount((FiniteAlphabet)count.getAlphabet());
		c.setCounts(count);
		addDistributionToCount(c, background);
		return DistributionTools.countToDistribution(c);
	}



	/**
	 * Adds the weights from the distribution to the counts from the given count. 
	 * The count and distribution have to be under the same alphabet.
	 * This method changes the first parameter (which is returned as a result).
	 * 
	 * @param count
	 * @param background
	 * @return
	 * @throws BioException
	 */
	private static Count addDistributionToCount(Count count, Distribution background) throws BioException{

		FiniteAlphabet a = (FiniteAlphabet) count.getAlphabet();
		if (!a.equals(background.getAlphabet()))
			throw new BioException("Alphabets of Count and Distribution are not equal (" + a.getName() + " vs. " + background.getAlphabet().getName() + ")");
		Iterator<?> iter = a.iterator();

		while (iter.hasNext()){
			AtomicSymbol s = (AtomicSymbol) iter.next();
			count.increaseCount(s, background.getWeight(s));
		}

		return count;
	}



	/**
	 * Counts relative entropy D(p||q) for two distributions.
	 * Relative entropy is always non-negative, and zero for identical distributions.
	 * This function is NOT symmetric, therefore it cannot be used as a metric.
	 * 
	 * @param p - first distribution
	 * @param q - second distribution
	 * @return
	 */
	public static double relativeEntropy(Distribution p, Distribution q) {

		double result = 0;
		if (distsEqual(p, q))
			return result;
		FiniteAlphabet a = (FiniteAlphabet ) p.getAlphabet();
		Iterator<?> iter = a.iterator();

		while (iter.hasNext()){
			Symbol s = (Symbol) iter.next();
			double ps;
			double qs;
			try {
				ps = p.getWeight(s);
			} catch (IllegalSymbolException e) {
				throw new MMFRuntimeException("Symbol " + s.getName() + " not found in Distribution under alphabet " + p.getAlphabet().getName(), e);
			}
			try {
				qs = q.getWeight(s);
			} catch (IllegalSymbolException e) {
				throw new MMFRuntimeException("Symbol " + s.getName() + " not found in Distribution under alphabet " + q.getAlphabet().getName(), e);
			}
			// TO AVOID PROBLEMS WITH 'MINUS ZERO"
			if (Math.abs(qs - ps) > EPSILON)
				result += ps * Tools.log(ps / qs, 2);
		}
		if (result < 0)
			throw new MMFRuntimeException("Relative entropy has to be >0, but was: " + result);
		return result;
	}



	public static Distribution averageDNADist(Distribution p, Distribution q) throws IllegalAlphabetException, IllegalSymbolException, ChangeVetoException{
		
		if (!p.getAlphabet().equals(dna))
			throw new IllegalAlphabetException("Required DNA, found alphabet: " + p.getAlphabet());
		if (!q.getAlphabet().equals(dna))
			throw new IllegalAlphabetException("Required DNA, found alphabet: " + q.getAlphabet());
		
		Distribution result = DistributionFactory.DefaultDistributionFactory.DEFAULT.createDistribution(dna);
		for (AtomicSymbol s: dnaSymbols){
			result.setWeight(s, (p.getWeight(s) + q.getWeight(s))/2);	
		}
		return result;
	}
	
	
	
	/**
	 * Counts the Dpq distance between two distributions. The formula is very simple and
	 * uses the notion of the relative entropy: 
	 * 	Dpq(P,Q) = sqrt (D(P||R) + D(Q||R)),
	 * where 
	 * 	R = (P + Q)/2 
	 *  D(P||Q) = relative entropy for distributions P and Q
	 *  
	 * @param p - first distribution
	 * @param q - second distribution
	 * @return
	 */
	public static double DpqDistance(Distribution p, Distribution q){
		
		if (distsEqual(p, q)) return 0;
		Distribution r;
		
		try {
			//Distribution dists[] = {p, q};
			//DistributionTools.average(dists);
			r = averageDNADist(p, q);
		} catch (BioException e) {
			throw new MMFRuntimeException(e);
		}
		
		double DpqSquare = relativeEntropy(p, r) + relativeEntropy(q, r);
		double result = Math.sqrt(DpqSquare);
		return result;
	}
	
	
	
	/**
	 * Read all sequences from a given FASTA file and return them in a map.
	 * 
	 * @param filename - path to the file (FASTA format) with sequences
	 * @param alphabetType - "DNA" or "PROTEIN" (also "RNA" will work)
	 * @return map with sequences' names as keys and sequences as values
	 * @throws FileNotFoundException
	 * @throws BioException 
	 * @throws NoSuchElementException 
	 */
	public static Map<String, RichSequence> readFastaSequences(String filename, String alphabetType) throws FileNotFoundException, NoSuchElementException, BioException {

		BufferedReader br = new BufferedReader(new FileReader(filename));
		RichSequenceIterator iter = RichSequence.IOTools.readFastaDNA(br,null);

		Map<String, RichSequence> result = new HashMap<String, RichSequence>();

		while (iter.hasNext()){
			RichSequence seq = iter.nextRichSequence();
			result.put(seq.getName(), seq);
			Logger.debug("Sequence read: %s (length: %d).\n", seq.getName(), seq.length());
		}
		return result;
	}


	/**
	 * Creates a simple FASTA file with sequences. Only name of the sequence in the first line,
	 * each sequence written in one line (no blanks, no newlines). This is required by BioProspector,
	 * Weeder erroneously reads file with blanks between nucleotides, and other motif finders may 
	 * need it as well.
	 *  
	 * @param path - prefix of the created file
	 * @param suffix - suffix of the created file
	 * @param seqMap - map of the sequences that shall be written to the file
	 * @return
	 * @throws MMFException
	 */
	public static File prepareSimplifiedFastaFile(String path, String suffix, Map<String, RichSequence> seqMap) throws MMFException 
	{
		File file;
		BufferedWriter writer = null;

		try {
			//file = File.createTempFile(prefix, suffix);
			file = new File(path + suffix);
			if (file.exists())
				file.delete();
			if (!file.createNewFile())
				throw new MMFRuntimeException("Failed to create temporary file with path " + path + " and suffix " + suffix);
			Logger.info("Created temporary file: %s\n", file.getAbsolutePath());
		} catch (IOException e) {
			throw new MMFRuntimeException("Failed to create temporary file with path " + path + " and suffix " + suffix);
		}

		try {
			writer = new BufferedWriter(new FileWriter(file));
			Iterator<String> seqIter  = seqMap.keySet().iterator();
			while (seqIter.hasNext()){
				RichSequence seq = seqMap.get(seqIter.next());
				Logger.debug("Writing next sequence: %s\n", seq.getName());
				writer.write('>' + seq.getName() + '\n');
				writer.write(seq.seqString() + '\n');
			}
		} catch (Exception e) {
			throw new MMFException(e);
		} finally {
			if (writer != null)
				try { writer.close();} catch (IOException e) {}
		}
		return file;
	}
	
	

	public static String toString(Distribution[] dists){

		StringBuffer buff = new StringBuffer();
		buff.append("     A     C     G     T\n");

		Formatter formatter = new Formatter(buff);
		int pos = 0;
		for (Distribution d: dists){
			pos ++;
			formatter.format("%2d.  ", pos);
			try {
				formatter.format("%.3f ", d.getWeight(a));
				formatter.format("%.3f ", d.getWeight(c));
				formatter.format("%.3f ", d.getWeight(g));
				formatter.format("%.3f ", d.getWeight(t));
			} catch (IllegalSymbolException e) {
				throw new MMFRuntimeException(e);
			}
			buff.append("\n");
		}
		buff.append("\n");

		return buff.toString();
	}


	public static String toString(Count[] counts){

		StringBuffer buff = new StringBuffer();
		buff.append("\tA\tC\tG\tT\t\n");

		Formatter formatter = new Formatter(buff);
		int pos = 0;
		for (Count co: counts){
			pos ++;
			buff.append(pos + ".\t");
			try {
				formatter.format("%.3f\t%.3f\t%.3f\t%.3f\t", 
						co.getCount(a),	co.getCount(c),	co.getCount(g),	co.getCount(t));
			} catch (IllegalSymbolException e) {
				throw new MMFRuntimeException(e);
			}
			buff.append("\n");
		}
		buff.append("\n");

		return buff.toString();
	}


	
	public static String toString(double[][] pspm){

		StringBuffer buff = new StringBuffer();
		buff.append("     A     C     G     T\n");

		Formatter formatter = new Formatter(buff);
		int pos = 0;
		for (double [] d: pspm){
			pos ++;
			buff.append(pos + ".  ");
			for (int nucl = 0; nucl < 4; nucl ++)
				formatter.format("%.3f ", d[nucl]);
			buff.append("\n");
		}
		buff.append("\n");
		return buff.toString();
	}
	
	
	public static boolean distsEqual(Distribution p, Distribution q){
		if (!p.getAlphabet().equals(q.getAlphabet()))
			return false;
		FiniteAlphabet a = (FiniteAlphabet) p.getAlphabet();
		Iterator<?> iter = a.iterator();
		while (iter.hasNext()){
			Symbol s = (Symbol) iter.next();
			try {
				if (p.getWeight(s) != q.getWeight(s))
					return false;
			} catch (IllegalSymbolException e) {
				throw new MMFRuntimeException(e);
			}
		}
		return true;
	}



	public static Distribution complement(Distribution d) throws WrongParameterException{

		try {
			FiniteAlphabet a = (FiniteAlphabet) d.getAlphabet();
			Distribution result = DistributionFactory.DefaultDistributionFactory.DEFAULT.createDistribution(a);

			if (a.equals(DNATools.getDNA()))
				for (AtomicSymbol s: dnaSymbols)
					result.setWeight(DNATools.complement(s), d.getWeight(s));
			else if (a.equals(RNATools.getRNA()))
				for (AtomicSymbol s: rnaSymbols)
					result.setWeight(RNATools.complement(s), d.getWeight(s));
			else 
				throw new WrongParameterException("Cannot reverse symbol from alphabet " + a.getName());

			return result;
		} catch (IllegalAlphabetException e) {
			throw new WrongParameterException(e);
		} catch (IllegalSymbolException e) {
			throw new WrongParameterException(e);
		} catch (ChangeVetoException e) {
			throw new MMFRuntimeException("Change veto caused by: " + e.getMessage(), e);
		}
	}



	public static Distribution[] reverseComplement(Distribution [] dists) throws WrongParameterException{

		int size = dists.length;
		Distribution [] result = new Distribution [size];
		Distribution d;

		for (int i = 0; i < size; i++){
			d = dists[size - i - 1];
			result[i] = complement(d);
		}

		return result;
	}



	public static SymbolList reverseComplement(SymbolList seq) throws WrongParameterException{
		try {
			if (seq.getAlphabet().equals(DNATools.getDNA()))
				return DNATools.reverseComplement(seq);
			else if (seq.getAlphabet().equals(RNATools.getRNA()))
				return RNATools.reverseComplement(seq);
		} catch (IllegalAlphabetException e) {
			throw new MMFRuntimeException("Unexpected problem: " + e.getMessage(), e);
		}
		throw new WrongParameterException("Cannot reverse sequence under alphabet " + seq.getAlphabet().getName());
	}



	public static RichSequence concatenate (List<RichSequence> sequences, String name, boolean addReverse) 
	throws WrongParameterException
	{

		if (sequences.size() < 1)
			throw new WrongParameterException("Cannot create a sequence from an empty list");
		Alphabet a = sequences.get(0).getAlphabet();
		StringBuffer buf = new StringBuffer();

		for (RichSequence s: sequences){
			if (!s.getAlphabet().equals(a))
				throw new WrongParameterException("Cannot concatenate sequences of different alphabets");
			buf.append(s.seqString());
		}

		if (addReverse){
			List<RichSequence> reversedList = new LinkedList<RichSequence>(sequences);
			Collections.reverse(reversedList);
			for (RichSequence s: reversedList){
				buf.append(reverseComplement(s).seqString());
			}	
		}

		try {
			RichSequence result = RichSequence.Tools.createRichSequence(name, buf.toString(), a);
			return result;
		} catch (BioException e) {
			throw new WrongParameterException("Failed to concatenate the given sequences because of: \n" + e.getMessage(), e);
		}
	}



//	TODO: this can't stay encoded like that...
	@Deprecated
	private static char getMainSymbol(Distribution d){
		FiniteAlphabet a  = (FiniteAlphabet) d.getAlphabet();
		Iterator<?> iter = a.iterator();
		char result = 0;
		double p = 0;
		while (iter.hasNext()){
			AtomicSymbol s = (AtomicSymbol) iter.next();
			try {
				if (d.getWeight(s) > p){
					p = d.getWeight(s);
					result = asChar(s);
				}
			} catch (IllegalSymbolException e) {
				throw new MMFRuntimeException(e);
			}
		}
		return result;
	}

	@Deprecated
	public static void addWeigthsToMotifs(List<Motif> motifs){
		int count = motifs.size();
		for (int position = 0; position < count; position++){
			double weight = ((double)count/(count + position)) + 0.5;
			motifs.get(position).setWeight(weight);
		}
	}
	
	
	public static String getConsensusSequence(Distribution[] d){
		StringBuffer buf = new StringBuffer();
		for (Distribution dis : d){
			buf.append(getMainSymbol(dis));
		}
		return buf.toString();
	}
	
	public static char asChar(AtomicSymbol s){
		return s.getName().charAt(0);
	}
}