package mimuw.mmf.util;

import static mimuw.mmf.util.BioTools.getDNACount;
import static mimuw.mmf.util.BioTools.getDNADistribution;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import mimuw.mmf.WrongParameterException;

import org.biojava.bio.BioException;
import org.biojava.bio.dist.Count;
import org.biojava.bio.dist.Distribution;
import org.biojava.bio.dist.IndexedCount;
import org.biojava.bio.dist.UniformDistribution;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.symbol.FiniteAlphabet;
import org.biojava.bio.symbol.IllegalAlphabetException;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.SymbolList;
import org.biojava.utils.ChangeVetoException;
import org.biojavax.bio.seq.RichSequence;
import org.junit.Before;
import org.junit.Test;

@SuppressWarnings("deprecation")
public class BioToolsTest {

	private FiniteAlphabet dnaAlphabet;
	Distribution uniformDNABackground;
	
	@Before
	public void setUp(){
		dnaAlphabet = DNATools.getDNA();
		uniformDNABackground = new UniformDistribution(dnaAlphabet);
	}

	
	

	
	

	@Test
	public void testGetDistance() {
		Distribution first0 = getDNADistribution(0.5, 0.1, 0.3, 0.1);
		Distribution first1 = uniformDNABackground;
		Distribution first2 = getDNADistribution(0.1, 0.7, 0.1, 0.1);
		Distribution first3 = getDNADistribution(0.3, 0.06, 0.6, 0.04);
		Distribution first4 = getDNADistribution(0.14, 0.26, 0.5, 0.1);
		Distribution first5 = getDNADistribution(0.34, 0.56, 0.06, 0.04);
		
		
		Distribution second0 = uniformDNABackground;
		Distribution second1 = getDNADistribution(0.15, 0.61, 0.08, 0.16);
		Distribution second2 = getDNADistribution(0.3, 0.05, 0.57, 0.08);
		
		Distribution first [] = {first0, first1, first2, first3, first4, first5};
		Distribution second[] = {second0, second1, second2};
		
		BioTools.DistributionDistanceCounter counter = new BioTools.RelativeEntropyCounter();
		BioTools.getBestAvgDistance(counter, first, second, uniformDNABackground);
	}
	
	
	@Test
	public void testGetBestAvgDistance() throws BioException, ChangeVetoException{
		
		Distribution first0 = getDNADistribution(0.5, 0.1, 0.3, 0.1);
		Distribution first1 = uniformDNABackground;
		
		Distribution second0 = getDNADistribution(0.3, 0.2, 0.3, 0.2);
		Distribution second1 = getDNADistribution(0.45, 0.11, 0.32, 0.12);
		Distribution second2 = uniformDNABackground;
		
		Distribution first [] = {first0, first1};
		Distribution second[] = {second0, second1, second2};
		
		BioTools.DistributionDistanceCounter counter = new BioTools.RelativeEntropyCounter();
		
		double result = BioTools.getBestAvgDistance(counter, first, second, uniformDNABackground);
		double expected = BioTools.getTotalDistance(counter, first, 1, second, uniformDNABackground);
		expected = BioTools.applyPenalty(expected/3, 2, 1);
		assertEquals(expected, result);
	}
	
	
	private void testGetTotalDistance(BioTools.DistributionDistanceCounter counter) throws BioException {

		Distribution first0 = getDNADistribution(0.5, 0.1, 0.3, 0.1);
		Distribution first1 = uniformDNABackground;
		
		Distribution second0 =  getDNADistribution(0.3, 0.2, 0.3, 0.2);
		Distribution second1 =  getDNADistribution(0.1, 0.4, 0.2, 0.3);
		Distribution second2 = second0;
		
		Distribution first [] = {first0, first1};
		Distribution second[] = {second0, second1, second2};
		
		double result = BioTools.getTotalDistance(counter, first, -1, second, uniformDNABackground);
		double expected = 
			counter.distance(first0, uniformDNABackground) 
			+ counter.distance(first1, second0)
			+ counter.distance(uniformDNABackground, second1)
			+ counter.distance(uniformDNABackground, second2);
	
		assertEquals(expected, result);
	}
	
	
	
	@Test
	public void testGetTotalDistanceRelativeEntropy() throws BioException {
		BioTools.DistributionDistanceCounter relEntCounter = new BioTools.RelativeEntropyCounter();
		this.testGetTotalDistance(relEntCounter);
	}
	
	
	
	@Test
	public void testGetTotalDistanceDpq() throws BioException {
		BioTools.DistributionDistanceCounter DpqCounter = new BioTools.DpqCounter();
		this.testGetTotalDistance(DpqCounter);
	}
	
	
	
	@Test
	public void testApplyLaplaceCorrection() throws BioException {

		Count count = new IndexedCount(dnaAlphabet);
		
		double a = 3.0;
		double c = 3.0;
		double g = 0.0;
		double t = 4.0;
		double sum = a + c + g + t;
		double sum1 = ++sum;
		
		count.setCount(DNATools.a(), a);
		count.setCount(DNATools.c(), c);
		count.setCount(DNATools.g(), g);
		count.setCount(DNATools.t(), t);

		Distribution corrected = BioTools.applyLaplaceCorrection(count, uniformDNABackground);
		
		assertEquals((a+0.25)/sum1, corrected.getWeight(DNATools.a()));
		assertEquals((c+0.25)/sum1, corrected.getWeight(DNATools.c()));
		assertEquals((g+0.25)/sum1, corrected.getWeight(DNATools.g()));
		assertEquals((t+0.25)/sum1, corrected.getWeight(DNATools.t()));
	}
	
	
	@Test
	public void testRelativeEntropyForSameDistributions() throws IllegalSymbolException, ChangeVetoException, IllegalAlphabetException{

		Distribution p =  getDNADistribution(0.5, 0.1, 0.3, 0.1);
		double relEnt = BioTools.relativeEntropy(p, p);
		Logger.debug("Relative entropy: %.5f\n", relEnt);
		assertEquals(0.0, relEnt);
	}

	
	@Test
	public void testRelativeEntropyForDifferentDistributions() throws IllegalSymbolException, ChangeVetoException, IllegalAlphabetException{

		Distribution p =  getDNADistribution(0.5, 0.1, 0.3, 0.1);
		Distribution q =  getDNADistribution(0.2, 0.2, 0.2, 0.4);
		
		double expected = 
			  0.5 * Tools.log(0.5 / 0.2, 2)
			+ 0.1 * Tools.log(0.1 / 0.2, 2)
			+ 0.3 * Tools.log(0.3 / 0.2, 2)
			+ 0.1 * Tools.log(0.1 / 0.4, 2);
		
		double relEnt = BioTools.relativeEntropy(p, q);
		Logger.debug("Relative entropy: %.5f\n", relEnt);
		assertEquals((float)expected, (float)relEnt);
	}
	
	@Test
	public void testDpqSymmetry(){
		Distribution p =  getDNADistribution(0.5, 0.1, 0.3, 0.1);
		Distribution q =  getDNADistribution(0.2, 0.2, 0.2, 0.4);
		double pq = BioTools.DpqDistance(p, q);
		double qp = BioTools.DpqDistance(q, p);
		assertEquals(pq, qp);
	}
	
	@Test
	public void testRelativeEntropyNoSymmetry(){
		Distribution p =  getDNADistribution(0.5, 0.1, 0.3, 0.1);
		Distribution q =  getDNADistribution(0.2, 0.2, 0.2, 0.4);
		double pq = BioTools.relativeEntropy(p, q);
		double qp = BioTools.relativeEntropy(q, p);
		assertFalse(pq == qp);
	}
	
	@Test
	public void testMotifThatFailed() throws BioException{
		final int LEN = 10;
		Count counts[] = new Count[LEN];
		counts[0] = getDNACount(0, 5, 1, 0);
		counts[1] = getDNACount(4, 1, 0, 1);
		counts[2] = getDNACount(4, 0, 0, 2);
		counts[3] = getDNACount(1, 1, 4, 0);
		counts[4] = getDNACount(2, 0, 1, 3);
		counts[5] = getDNACount(2, 3, 1, 0);
		counts[6] = getDNACount(3, 2, 0, 1);
		counts[7] = getDNACount(1, 1, 4, 0);
		counts[8] = getDNACount(3, 0, 3, 0);
		counts[9] = getDNACount(0, 3, 0, 3);
		
		Distribution[] dists = new Distribution [LEN];
		for (int i = 0; i < counts.length; i++){
			dists[i] = BioTools.applyLaplaceCorrection(counts[i], uniformDNABackground);
		}
		BioTools.DistributionDistanceCounter counter1 = new BioTools.DpqCounter();
		//BioTools.DistributionDistanceCounter counter2 = new BioTools.RelativeEntropyCounter();
		
		double dist = BioTools.getBestAvgDistance(counter1, dists, dists, uniformDNABackground);
		assertEquals(0.0, dist);
	}
	
	

	
	@Test
	public void testReverseComplementSeq() throws BioException, WrongParameterException{
		RichSequence seqA = RichSequence.Tools.createRichSequence("seqA", "aAGTGCAcacc", dnaAlphabet);
		SymbolList rev = BioTools.reverseComplement(seqA);
		assertEquals(rev.seqString(), "ggtgtgcactt");
	}
	
	@SuppressWarnings("deprecation")
	@Test
	public void testReverseComplementDists() throws WrongParameterException{
		Distribution p[] =  {getDNADistribution(0.5, 0.1, 0.3, 0.1), getDNADistribution(0.3, 0.3, 0.3, 0.1)};
		Distribution pRev[] = {getDNADistribution(0.1, 0.3, 0.3, 0.3), getDNADistribution(0.1, 0.3, 0.1, 0.5)};
		assertEquals(pRev, BioTools.reverseComplement(p));
	}

}