package mimuw.mmf.util;

import static mimuw.mmf.util.BioTools.EPSILON;
import static mimuw.mmf.util.BioTools.a;
import static mimuw.mmf.util.BioTools.c;
import static mimuw.mmf.util.BioTools.dna;
import static mimuw.mmf.util.BioTools.g;
import static mimuw.mmf.util.BioTools.rna;
import static mimuw.mmf.util.BioTools.t;
import static mimuw.mmf.util.BioTools.u;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import mimuw.mmf.WrongParameterException;
import mimuw.mmf.clusterers.DistanceCounter;
import mimuw.mmf.finders.MMFRuntimeException;
import mimuw.mmf.motif.AbstractMotif;
import mimuw.mmf.motif.NotInitializedException;

import org.biojava.bio.BioException;
import org.biojava.bio.dist.Distribution;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.RNATools;
import org.biojava.bio.seq.Sequence;
import org.biojava.bio.symbol.AtomicSymbol;
import org.biojava.bio.symbol.FiniteAlphabet;
import org.biojava.bio.symbol.IllegalAlphabetException;
import org.biojava.bio.symbol.IllegalSymbolException;
import org.biojava.bio.symbol.Symbol;

/**
 * This class doubles many operations from BioTools class, but here implementations are more 
 * general, however much faster. Instead of working on BioJava classes such as Distribution
 * or Count, QuickBioTools generally uses java arrays. 
 * 
 * @author milosz
 *
 */
public class QuickBioTools {

	public static final int MINIMAL_OVERLAP = 1; 


	public static interface ColumnsDistributionComparator /*extends DistanceCounter<double[]>*/{
		public double distance(double p[], double q[]);
	}

	public static class CRECounter implements ColumnsDistributionComparator {
		public double distance(double p[], double q[]){
			return relativeEntropy(p, q, 0);		
		}
	}

	public static class CDpqCounter implements ColumnsDistributionComparator {
		public double distance(double p[], double q[]){
			return quickDpqDistance(p, q, 0);
		}
	}

	public static class CPCCounter implements ColumnsDistributionComparator {
		public double distance(double p[], double q[]){
			try {
				return PearsonsCorrelationToDistance(PearsonsCorrelation(p, q, 0));
			} catch (WrongParameterException e) {
				throw new MMFRuntimeException(e);
			}
		}
	}


	public static interface FittingOnSequenceComparator {
		public double distance(double d1[], double d2[], int d1ShiftRight);
	}

	public static class PPCCounter implements FittingOnSequenceComparator {
		public double distance(double[] d1, double[] d2, int d1ShiftRight) {
			try {
				return PearsonsCorrelationToDistance(QuickBioTools.PearsonsCorrelation(d1, d2, d1ShiftRight));
			} catch (WrongParameterException e) {
				throw new MMFRuntimeException(e);
			}
		}
	}

	public static class PRECounter implements FittingOnSequenceComparator {
		public double distance(double[] d1, double[] d2, int shiftRight) {
			return relativeEntropy(d1, d2, shiftRight);
		}
	}

	public static class PDpqCounter implements FittingOnSequenceComparator {
		public double distance(double[] d1, double[] d2, int shiftRight) {
			return quickDpqDistance(d1, d2, shiftRight);
		}
	}


	/**
	 * Correct, but slower than its successor
	 */
	@Deprecated
	protected static double DpqDistance(double p[], double q[]) throws WrongParameterException{

		if (p == q)	return 0;
		if (p.length != q.length)
			throw new WrongParameterException("Arrays of same length expected (got " + p.length + ", " + q.length + ")");

		double r[] = new double[p.length];
		for (int i =0; i < p.length; i++){
			r[i] = (p[i] + q[i]) / 2;
		}

		double DpqSquare = relativeEntropy(p, r, 0) + relativeEntropy(q, r, 0);
		return Math.sqrt(DpqSquare);
	}


	protected static double quickDpqDistance(double p[], double q[], int pRightShift){

		if (pRightShift == 0 && p == q)
			return 0;

		int pi=0, qi=0;
		double ps, qs, DpqSquare=0;

		if (pRightShift > 0)	
			qi = pRightShift;
		else
			pi = -pRightShift;

		while (pi < p.length && qi < q.length){	
			ps = p[pi];
			qs = q[qi];
			if (Math.abs(qs - ps) > EPSILON){ 		// TO AVOID PROBLEMS WITH 'MINUS ZERO"
				DpqSquare += ps * Tools.log((2*ps) / (ps+qs), 2);
				DpqSquare += qs * Tools.log((2*qs) / (ps+qs), 2);
			}
			pi++; qi++;
		}
		if (DpqSquare < 0)
			throw new MMFRuntimeException("Dpq has to be >0, but was: " + DpqSquare);
		return Math.sqrt(DpqSquare);
	}


	protected static double relativeEntropy(double p[], double q[], int pRightShift){

		if (pRightShift == 0 && p == q)
			return 0;

		int pi=0, qi=0;
		double ps, qs, resultHuge=0, resultTiny=0;

		if (pRightShift > 0)	
			qi = pRightShift;
		else
			pi = -pRightShift;

		while (pi < p.length && qi < q.length){	
			ps = p[pi];
			qs = q[qi];
			double current = ps * Tools.log(ps / qs, 2);
			//if (Math.abs(qs - ps) > EPSILON) 	// TO AVOID PROBLEMS WITH 'MINUS ZERO"
			if (current > 0.1)
				resultHuge += current;
			else 
				resultTiny += current;//result += 
			pi++; qi++;
		}
		double result = resultHuge + resultTiny;
		if (result < 0)
			throw new MMFRuntimeException("Relative entropy has to be >0, but was: " + result);
		return result;
	}


	
	private static double getSumOfColumnsDistributionDistances(
			ColumnsDistributionComparator counter, 
			double[][] first, 
			int firstShiftedRight, 
			double[][] second,
			double[] background)
	{
		double result = 0;
		int indexFirst = 0;
		int indexSecond = 0;
		int firstLen = first.length;
		int secondLen = second.length;

		if (-firstShiftedRight > firstLen)
			throw new MMFRuntimeException("Motif shifted too far left (" + firstShiftedRight + ")");
		if (firstShiftedRight > secondLen)
			throw new MMFRuntimeException("Motif shifted too far right (" + firstShiftedRight + ")");

		// START: counts sum of the distances between padding and the beginning
		if (firstShiftedRight < 0){
			for (; indexFirst < -firstShiftedRight; indexFirst++){
				result += counter.distance(first[indexFirst], background);
			}
		} else {
			for (; indexSecond < firstShiftedRight; indexSecond++){
				result += counter.distance(background, second[indexSecond]);
			}
		}

		// MIDDLE: counts the main part: distances of the corresponding double[]s
		while (indexFirst < firstLen && indexSecond < secondLen){
			result += counter.distance(first[indexFirst], second[indexSecond]);
			indexFirst++;
			indexSecond++;
		}

		// END: adds distances between the double[]s at the end and padding in the other sequence
		while (indexFirst < firstLen){
			result += counter.distance(first[indexFirst], background);
			indexFirst++;
		}
		while (indexSecond < secondLen){
			result += counter.distance(background, second[indexSecond]);
			indexSecond++;
		}
		return result;
	}
	
	
	
	/**
	 * Counts scores for all possible shifts (with at least one position overlapping), and returns the result
	 * as a pair: (<shift for best score>, <best score>)
	 *  
	 * @param counter this object will be used to count the distance between two motifs with each possible shift
	 * and using background as 'padding'
	 * @param first
	 * @param second
	 * @param background
	 * @return
	 */
	protected static Pair<Integer, Double> getBestAvgColumnsDistributionDistance (
			ColumnsDistributionComparator counter, 
			double[][] first, 
			double[][] second,
			double[] background)
			{
		double bestScore = Double.MAX_VALUE;
		int len1 = first.length;
		int len2 = second.length;
		int maxLeftShift  = (len1 > MINIMAL_OVERLAP) ? (len1 - MINIMAL_OVERLAP) : 0;
		int maxRightShift = (len2 > MINIMAL_OVERLAP) ? (len2 - MINIMAL_OVERLAP) : 0;
		int bestShift = maxLeftShift;

		for (int firstShiftedRight = -maxLeftShift; firstShiftedRight <= maxRightShift; firstShiftedRight ++){

			int overlapLength = firstShiftedRight < 0 ? Math.min(len1+firstShiftedRight, len2)	: Math.min(len1, len2 - firstShiftedRight);
			int outsideLength = len1 + len2 - 2*overlapLength;
			//int totalLength = outsideLength + overlapLength;

			double current = getSumOfColumnsDistributionDistances(counter, first, firstShiftedRight, second, background);
			double average = current / (overlapLength + outsideLength);
			double withPenalty = average;// TODO - decyzja BioTools.applyPenalty(average, overlapLength, outsideLength);

			if (withPenalty < bestScore){
				bestScore = withPenalty;
				bestShift = firstShiftedRight;
			}				
		}
		return new Pair<Integer, Double>(bestShift, bestScore);
			}



	public static double getColumnsDistributionDistance(
			AbstractMotif m1, 
			AbstractMotif m2, 
			ColumnsDistributionComparator counter, 
			Distribution background)
	{
		try {
			double[][] d1 = m1.getPSPM();
			double[][] d2 = m2.getPSPM();
			double[][] d2Rev = m2.getRevCompPSPM();
			double[] bg = toQuickArray(background);

			Pair<Integer, Double> d1ToD2 = getBestAvgColumnsDistributionDistance(counter, d1, d2, bg);
			Pair<Integer, Double> d1ToD2Rev = getBestAvgColumnsDistributionDistance(counter, d1, d2Rev, bg);
			Logger.debug("   FWD Shift: %2d %c (%.2f)  REVCOMP Shift: %2d %c, Score: %.2f",
					Math.abs(d1ToD2.first()), 
					d1ToD2.first() < 0 ? 'L' : 'R', d1ToD2.second(), 
					Math.abs(d1ToD2Rev.first()), 
					d1ToD2Rev.first() < 0 ? 'L' : 'R', d1ToD2Rev.second());
			return Math.min(d1ToD2.second(), d1ToD2Rev.second());
		} catch (BioException e) {
			throw new MMFRuntimeException(e);
		} finally {

		}
	}



	public static double getFittingOnSequenceDistance (AbstractMotif m1, AbstractMotif m2, FittingOnSequenceComparator comparator) 
	throws  WrongParameterException, NotInitializedException 
	{
		/*int maxLeftShift = (m1.length() > MINIMAL_OVERLAP) ? m1.length() - MINIMAL_OVERLAP : 0;
		int maxRightShift = (m2.length() > MINIMAL_OVERLAP) ? (m2.length() - MINIMAL_OVERLAP) : 0;*/
		
		/*
		 * The code below is not intuitive. 
		 * maxRightShift stands for the maximal number of positions, that we could shift the first motif's 
		 * distribution to the right. If we were to compare motif positions, we would have to swap 
		 * maxRightShift with maxLeftShift, because if the first one is shifted by one position 
		 * to the right against the second one, its distribution of fitting on the sequence is 
		 * actually shifted by one position to the left against the second one's. 
		 */
		
		int maxRightShift = (m1.length() > MINIMAL_OVERLAP) ? (m1.length() - MINIMAL_OVERLAP) : 0;
		int maxLeftShift = (m2.length() > MINIMAL_OVERLAP) ? m2.length() - MINIMAL_OVERLAP : 0;
		
		double d1ToD2 = getBestFittingResult(m1.fittingProbabilities(), 
				m1.getSumOfFittingProbabilities(),
				m2.fittingProbabilities(), 
				m2.getSumOfFittingProbabilities(), 
				maxLeftShift, maxRightShift, comparator);
		double d1ToD2Rev = getBestFittingResult(m1.fittingProbabilities(), 
				m1.getSumOfFittingProbabilities(),
				m2.getFittingProbabilitiesRevComp(), 
				m2.getSumOfFittingProbabilitiesRevComp(), 
				maxLeftShift, maxRightShift, comparator);
		return Math.min(d1ToD2, d1ToD2Rev);
	}


	
	public static double[] normalize (double dist[], int from, int to) throws WrongParameterException {
		int len = dist.length;
		if (from < 0 || from >= len)
			throw new WrongParameterException("Parameter 'from' value: " +  from +", size of double[]: " + len);
		if (to < 0 || to >= len)
			throw new WrongParameterException("Parameter 'to' value: " +  to +", size of double[]: " + len);
		if (from > to)
			throw new WrongParameterException("Parameter 'from' value: " +  from +", 'to': " + to);
		
		double sum = 0;
		double result[] = new double[to-from+1];
		for (int i = from; i <= to; i++){
			sum += dist[i];
		}
		//System.out.printf("(DELETE) Suma na pozycjach %d-%d: %.10f\n", from, to, sum);
		for (int i = from; i <= to; i++){
			result[i-from] = dist[i]/sum;
		}
		return result;
	}


	public static double[] normalize (double dist[], double distSum, int from, int to) throws WrongParameterException {
		int len = dist.length;
		if (from < 0 || from >= len)
			throw new WrongParameterException("Parameter 'from' value: " +  from +", size of double[]: " + len);
		if (to < 0 || to >= len)
			throw new WrongParameterException("Parameter 'to' value: " +  to +", size of double[]: " + len);
		if (from > to)
			throw new WrongParameterException("Parameter 'from' value: " +  from +", 'to': " + to);
		
		double result[] = new double[to-from+1];
		
		for (int i = 0; i < from; i++){
			distSum -= dist[i];
		}
		for (int i = to+1; i < dist.length; i++){
			distSum -= dist[i];
		}
		//System.out.printf("(DELETE) Suma na pozycjach %d-%d: %.10f\n", from, to, sum);
		for (int i = from; i <= to; i++){
			result[i-from] = dist[i]/distSum;
		}
		return result;
	}
	
	
	private static double getBestFittingResult (
			double[] d1,
			double d1Sum,
			double[] d2, 
			double d2Sum,
			int maxShiftFirstLeft,
			int maxShiftFirstRight,
			FittingOnSequenceComparator comparator) throws WrongParameterException 
	{
		
		int bestShift = 0;
		int len1 = d1.length;
		int len2 = d2.length;
		double current, result = Double.MAX_VALUE;
		
		for (int firstShiftedRight = -maxShiftFirstLeft; firstShiftedRight <= maxShiftFirstRight; firstShiftedRight++){
			int from1 = (firstShiftedRight <= 0) ? -firstShiftedRight : 0;
			int from2 = (firstShiftedRight <= 0) ? 0 : firstShiftedRight;
			int to1;
			int to2;
			if (firstShiftedRight <= 0){
				to1 = (len2 >= len1 + firstShiftedRight) ? (len1 - 1) : (len2 - firstShiftedRight - 1);
				to2 = (len2 >= len1 + firstShiftedRight) ? (len1 - 1 + firstShiftedRight) : (len2 - 1);
			} else {
				to1 = (len2 >= len1 - firstShiftedRight) ? (len1 - 1) : (len2 - firstShiftedRight - 1);
				to2 = (len2 >= len1 - firstShiftedRight) ? (len1 - 1 - firstShiftedRight) : (len2 - 1);
			}
			//double d1Cut[] = normalize(d1,  from1, to1);
			//double d2Cut[] = normalize(d2,  from2, to2);;
			
			
			double d1Cut[] = normalize(d1, d1Sum, from1, to1);
			double d2Cut[] = normalize(d2, d2Sum, from2, to2);;
			//current = comparator.distance(d1, d2, firstShiftedRight);
			current = comparator.distance(d1Cut, d2Cut, 0);
			//Logger.info("\t\tShift first fitting: %2d (Score: %.4f, overlap: %d)\n", firstShiftedRight, current, d1Cut.length);
			if (current < result){
				result = current;
				bestShift = firstShiftedRight;
			}
		}
		/*
		 *  The explanation for (bestShift>=0?'L':'R'):
		 *  If the distance is smallest when the first distribution is shifted right relatively to 
		 *  the second, the first motif is actually shifted LEFT relatively to the second one.
		 *  And of course it works both ways.
		 */
		Logger.debug("\tShift first: %2d %c (Score: %.4f)", Math.abs(bestShift), (bestShift>=0?'L':'R'), result);
		return result;
	}



	protected static double PearsonsCorrelationToDistance(double p) throws WrongParameterException{
		if (Math.abs(p) > (1 + EPSILON))
			throw new WrongParameterException("Parameter has to be between -1 and 1, was: " + p);
		return (1-p)/2;
	}




	protected static double PearsonsCorrelation(double d1[], double d2[], int d1ShiftRight){

		double sumX=0, sumY=0, sumXY = 0, sumX2=0, sumY2=0;
		int x, y;

		if (d1ShiftRight > 0){
			x = 0; 
			y = d1ShiftRight;
		} else {
			x = -d1ShiftRight;
			y = 0;
		}

		while (x < d1.length && y < d2.length){	
			sumX += d1[x];
			sumY += d2[y];
			sumXY += d1[x]*d2[y];
			sumX2 += d1[x]*d1[x];
			sumY2 += d2[y]*d2[y];
			y++;
			x++;
		}

		int n = Math.min(x, y);
		double numerator = n*sumXY - sumX*sumY;
		double denominator = Math.sqrt((n*sumX2 - sumX*sumX) * (n*sumY2 - sumY*sumY));
		return numerator/denominator;
	}



	public static double[] getMotifProbsOnSingleSequence(double motif[][], Sequence seq) 
	throws WrongParameterException
	{

		int diff = seq.length() - motif.length;
		if (diff < 0)
			throw new WrongParameterException("Motif longer than the sequence");
		double result [] = new double[diff + 1];

		for (int posInSeq = 1; posInSeq <= diff + 1; posInSeq ++){
			result[posInSeq-1] = countMotifsProbabilityAtPosition(motif, seq, posInSeq);
			//System.out.printf("P[%d] = %.8f\n", posInSeq-1, result[posInSeq-1]);
		}
		return result;
	}



	/**
	 * Counts the probability that this motif appears on exactly this position in given sequence
	 * assuming that motif is described by the rectangular array of doubles. 
	 * 
	 */
	private static double countMotifsProbabilityAtPosition(double[][] motif, Sequence seq, Integer posInSeq){

		if (seq.length() < (posInSeq-1) + motif.length || posInSeq < 1)
			throw new MMFRuntimeException("Cannot apply given motif on position " + posInSeq + " of sequence " + seq.getName());

		double result = 1;
		int last = posInSeq + motif.length - 1;
		int posInMotif = 0;

		for (; posInSeq <= last; posInSeq++){
			Symbol symbol = seq.symbolAt(posInSeq);
			double localProb = motif[posInMotif++][nucleotideToPos(symbol)];
			result *= localProb;
			//System.out.printf("Probability at position %d: *%.4f(%s) = %f\n",	posInSeq, localProb, seq.symbolAt(posInSeq).getName().charAt(0), result);
		}

		return result;
	}	



	private static int nucleotideToPos(Symbol s){
		if (s.equals(a)) return 0;
		if (s.equals(c)) return 1;
		if (s.equals(g)) return 2;
		if (s.equals(t) || s.equals(u))	return 3;
		throw new MMFRuntimeException("Unrecognized symbol: " + s.getName());
	}



	private static double[] toQuickArray(Distribution d) throws IllegalAlphabetException, IllegalSymbolException{

		double result[] = new double[4];
		AtomicSymbol [] symbols;

		if (d.getAlphabet().equals(dna)) 		symbols = BioTools.dnaSymbols;
		else if (d.getAlphabet().equals(rna))	symbols = BioTools.rnaSymbols;
		else throw new IllegalAlphabetException("Unsupported alphabet: " + d.getAlphabet().getName());

		for (AtomicSymbol s: symbols)
			result[nucleotideToPos(s)] = d.getWeight(s);

		return result;
	}



	public static double[][] toPSPMArray(Distribution d[]) throws IllegalAlphabetException, IllegalSymbolException{

		double[][] quickDistribution = new double[d.length][4];

		for (int i = 0; i < d.length; i++){
			quickDistribution[i] = toQuickArray(d[i]);
		}
		return quickDistribution;
	}



	public static double[][] toPSPMArrayRevComp(Distribution d[]) throws IllegalSymbolException{

		double[][] quickRevCompDistribution = new double[d.length][4];

		for (int i = 0; i < d.length; i++){
			double [] revPos = quickRevCompDistribution[d.length - i - 1];
			if (d[i].getAlphabet().equals(DNATools.getDNA()) )
				revPos[0] = d[i].getWeight(DNATools.t());
			else 
				revPos[0] = d[i].getWeight(RNATools.u());
			revPos[1] = d[i].getWeight(DNATools.g());
			revPos[2] = d[i].getWeight(DNATools.c());
			revPos[3] = d[i].getWeight(DNATools.a());
		}
		return quickRevCompDistribution;
	}



	public static double getIC(double dist[], double background[]){
		if (background.length != dist.length)
			throw new MMFRuntimeException("Different lengths of arrays - impossible to count the IC");
		double result = 0;
		for (int pos = 0; pos < background.length; pos ++){
			double freq = dist[pos];
			if (freq == 0)
				continue;
			result += (freq*Tools.log(freq/background[pos], 2) );
		}
		return result;
	}
	
	
	
	public static double getAvgIC(double d[][], double bg[]){
		double sum = 0;
		for (int i = 0; i < d.length; i++)
			sum += getIC(d[i], bg);
		return sum/d.length;
	}
	
	
	/**
	 * Converts list of objects to the sorted list, where each element is a pair that consists of 
	 * compared objects and the distance between them. The list is sorted according to the distance.
	 * This method is implemented especially for checking, what the distances between motifs from 
	 * jaspar are.
	 * @param <T> type of objects, e.g. AbstractMotif
	 * @param objects
	 * @param distanceCounter
	 * @return
	 */
	public static <T> List<Pair<Pair<T, T>, Double>> sort(
			List<? extends T> objects, DistanceCounter<T> distanceCounter){

		int size = objects.size();
		double distances[][] = new double[size][size];

		List<Pair<Pair<T, T>, Double>> list = new ArrayList<Pair<Pair<T,T>,Double>>();

		for (int i = 0; i < size; i++){
			T obj1 = objects.get(i);
			Logger.info("Comparing %s to:\n", obj1.toString());
			for (int j = 0; j < size; j++){
				T obj2 = objects.get(j);
				double dist = distanceCounter.distance(obj1, obj2);
				//double dist = (obj1 == obj2) ? 0 : distanceCounter.distance(obj1, obj2);
				Logger.info("  to %s\n", obj2.toString());
				distances[i][j] = dist;
				if (i!= j)
					list.add(new Pair<Pair<T, T>, Double>(new Pair<T,T>(obj1, obj2), dist));
			}
		}
		
		Collections.sort(list, 
			new Comparator<Pair<Pair<T, T>, Double>>(){
				public int compare(Pair<Pair<T, T>, Double> o1, Pair<Pair<T, T>, Double> o2) {
					if (o1.second() > o2.second()) return 1;
					if (o1.second() < o2.second()) return -1;
					return 0;
				}
			}
		);
		return list;
	}


	
	public static String getConsensusSequence(double[][] pspm){
		StringBuffer buf = new StringBuffer();
		for (double [] dis: pspm){
			buf.append(getMainSymbol(dis));
		}
		return buf.toString();
	}

	private static char getMainSymbol(double[] d){
		FiniteAlphabet a  = dna;
		Iterator<?> iter = a.iterator();
		char result = 'n';
		double p = 0;
		while (iter.hasNext()){
			AtomicSymbol s = (AtomicSymbol) iter.next();
			int pos = nucleotideToPos(s);
			if (d[pos] > p){
				p = d[pos];
				result = BioTools.asChar(s);
			}
		}
		return result;
	}

	
	protected static String getRepresentation (double [] dist, int size){
		
		StringBuffer buf = new StringBuffer();
		int [] counts = toCounts(dist, size);
		
		for (AtomicSymbol s : BioTools.dnaSymbols){
			int pos = nucleotideToPos(s);
			for (int i = 0; i < counts[pos]; i++){
				buf.append(BioTools.asChar(s));
			}
		}
		return buf.toString();
	}
	
	private static int[] toCounts (double[] dist, int sum){
		double step = 1.0 / sum;
		dist = dist.clone();
		int result [] = new int[dist.length];
		int i = 0;
		while (i < sum){
			int greatest = findMax(dist);
			result[greatest]++;
			dist[greatest] -= step;
			i++;
		}
		return result;
	}
		
	
	private static int findMax(double [] arr){
		double greatest = Double.MIN_VALUE;
		int result = 0;
		for (int i = 0; i < arr.length; i++){
			if (arr[i] > greatest){
				result = i;
				greatest = arr[i];
			}
		}
		return result;
	}
		
	
	public static double[] getDistribution(double gcContent){
		double result[] = new double[4];
		result [nucleotideToPos(DNATools.a())] = (1-gcContent)/2;
		result [nucleotideToPos(DNATools.c())] = gcContent/2;
		result [nucleotideToPos(DNATools.g())] = gcContent/2;
		result [nucleotideToPos(DNATools.t())] = (1-gcContent)/2;
		return result;
	}
	
	
	
	public static double getMaxDistanceBetweenSpecifiedMotifs(List<AbstractMotif> motifs, List<List<Double>> distances, String origin){
		
		double result = Double.MAX_VALUE;
		
		for (int pos1 = 0; pos1 < motifs.size(); pos1++){
			AbstractMotif m1 = motifs.get(pos1);
			if (m1.origin().equals(origin)){
				for (int pos2 = 0; pos2 < motifs.size(); pos2++){
					AbstractMotif m2 = motifs.get(pos2);
					if (pos1 != pos2 && m2.origin().equals(origin)){
						result = Math.min(result, distances.get(pos1).get(pos2));
					}
				}
			}
		}	
		return result;
	}
}