package mimuw.mmf.util;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import mimuw.mmf.WrongParameterException;

/**
 * This class contains various static methods used in the application.
 * None of these methods concerns biological data, such methods can be found in 
 * mimuw.motifs.BioTools class
 * 
 * @author milosz
 *
 */
public class Tools {

	public static Random random = new Random(); 

	/**
	 * Reads the parameters file and returns the values as a map. The file will be interpreted as having the format described below.
	 * 
	 * FORMAT: each line has to begin with the program identifier (meme, mdscan, bioprospector) followed by some whitespace, 
	 * and then - all parameters delimited by sign '|' (vertical line)
	 * Comments are allowed, but only for the entire line. Such line has to begin with hash character (#)
	 * Example:
	 * #BEGIN
	 * #this is the parameters file
	 * meme  -w 7 | -mod oops | -pal
	 * mdscan -w 7 | -s 40
	 * bioprospector -W 7 | -T 5
	 * #END
	 * 
	 * @param filename - name of the text file with parameters
	 * @return
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	@Deprecated
	public static Map<String, String> readParametersFile(String filename) throws FileNotFoundException, IOException{

		Map<String, String> result = new HashMap<String, String>();
		BufferedReader bufRead = new BufferedReader(new FileReader(filename));

		String line;
		while ((line = bufRead.readLine()) != null){
			if (line.trim().startsWith("#"))	// comment 
				continue;
			if (line.trim().equals("")) 
				continue;
			String tokens[] = line.split("\\s", 2);
			String programIdentifier = tokens[0].toLowerCase();

			String paramStr = (tokens.length < 2) ? "" : tokens[1];
			result.put(programIdentifier, paramStr);
		}
		return result;
	}

	
	@Deprecated
	public static Map<String, String> splitParametersString(String s, String delim){

		String[] tokens = s.split(delim);
		Map<String, String> result = new HashMap<String, String>();

		for (String token: tokens){
			String keyVal[] = token.split("\\s", 2);
			String val = keyVal.length == 2 ? keyVal[1] : null;
			result.put(keyVal[0], val);
		}
		return result;
	}




	/**
	 * Prints all data from the given file to the given stream.
	 *  
	 * @param filename - file to read and get data from
	 * @param stream - stream where the data will be printed
	 * @throws IOException
	 */
	public static void writeFileToStream(String filename, PrintStream stream) throws IOException{
		BufferedReader bufRead = new BufferedReader(new FileReader(filename));
		String line;
		while ((line = bufRead.readLine()) != null)
			stream.println(line);
	}


	
	public static int executeProcessAndFlushStreams(Process proc) throws InterruptedException{
		StreamGobbler errorGobbler = new StreamGobbler(proc.getErrorStream(), "ERROR");
		StreamGobbler outputGobbler = new StreamGobbler(proc.getInputStream(), "OUTPUT");
		errorGobbler.start();
		outputGobbler.start();
		int exitVal = proc.waitFor();
		errorGobbler.join();
		outputGobbler.join();
		Logger.debug("ExitValue: %d\n", exitVal);
		return exitVal;
	}
	
	
	
	/**
	 * Return sublist of a given list, chosen randomly (each element has the same probability 
	 * to appear on the new list)
	 * 
	 * @param <T> - type of the elements contained by the list
	 * @param size - number of randomly chosen elements (has to be less than size of the list)
	 * @param objects - list of the objects
	 * @return
	 * @throws WrongParameterException
	 */
	public static <T> List<T> sublist(int size, List<T> objects) throws WrongParameterException{

		if (size > objects.size())
			throw new WrongParameterException("Input data set too small ("+objects.size()+", required randoms: " + size +")");
		
		List<T> o = new ArrayList<T>(objects);
		List<T> result = new ArrayList<T>();
		int index;
		Random rand = new Random();
		
		while (size > 0){
			index = rand.nextInt(o.size());
			result.add(o.get(index));
			o.remove(o.get(index));
			size --;
		}
		
		return result;
	}
	
	

	/**
	 * Returns true whether the given string can be converted into an integer belonging
	 * to the given compartment. Otherwise returns false.
	 * 
	 * @param value - string to analyse
	 * @param min - left boundary of the compartment
	 * @param max - right boundary of the compartment
	 * @return
	 */
	public static boolean isIntegerBetween(String value, int min, int max){
		try {
			int intval = new Integer(value).intValue();
			if (intval < min || intval > max)
				return false;
			return true;
		} catch (NumberFormatException e){
			return false;
		}
	}


	public static boolean isFloatBetween(String value, float min, float max){
		try {
			float fval = new Float(value);
			if (fval < min || fval > max)
				return false;
			return true;
		} catch (NumberFormatException e){
			return false;
		}
	}

	
	
	/**
	 * Checks whether the given string can be converted into a positive float value.
	 * If yes - returns true, otherwise false.
	 * 
	 * @param value - String to analyse
	 * @return
	 */
	public static boolean isPositiveFloat(String value){
		return isFloatBetween(value, (float) 0.0, Float.MAX_VALUE);
	}


	public static double log(double d, double base) {
		return Math.log(d)/Math.log(base);
	}

	/*public static double log2(double d) {
		return log(d, 2);
	}*/

	public static double log10(double d) {
		return log(d, 10);
	}
}